#include     <objbase.h>       //This program is exactly similiar to the CA COM object I developed in C++ in my 1st COM
#include     <stdio.h>         //Object Memory Tutorial.  But in this tutorial we are going a bit lower level and will
#include     "Registry.h"      //use plain C and plain PowerBASIC, i.e., PowerBASIC without Version 9's functionality.

const      CLSID CLSID_CB      ={0x20000000,0x0000,0x0000,{0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x10}}; 
const      IID   IID_IX        ={0x20000000,0x0000,0x0000,{0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x11}};
const      IID   IID_IY        ={0x20000000,0x0000,0x0000,{0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x12}};
HINSTANCE  g_hModule           = NULL;              //Store dll instance handle
const char g_szFriendlyName[]  = "Com Object CB";   //Store friendly name of component
const char g_szVerIndProgID[]  = "ComObject.CB";    //Store Version Independent ProgID
const char g_szProgID[]        = "ComObject.CB.1";  //Store Versioned Program ID.
long       g_lObjs=0;
long       g_lLocks=0;

typedef struct IXVtbl IXVtbl; //All these two lines do is specify the existance of a struct variable named IXVtbl or IYVtbl 
typedef struct IYVtbl IYVtbl; //respectively without actually defining what these structures are or allocating any storage for  

typedef interface IX          //same.  The typedef just left though does both define and allocate storage for another entity         
{                             //that contains a pointer to one of these as yet undefined entities, i.e., an IXVtbl or an IYVtbl.
 const IXVtbl* lpVtbl;        //C allows this because while it may not as yet know what an IXVtbl or IYVtbl is, it does know how 
}IX;                          //big a pointer to one of them would be.  The critical lines immediately below though finally 

typedef interface IY          //reveal just what an IXVtbl and an IYVtbl are.  We see that an IXVtbl and an IYVtbl contain 
{                             //function pointers.  So lets take stock at this point of just what has been defined only, and    
 const IYVtbl* lpVtbl;        //what has actualy been allocated.  Its not hard to do. We do know what an IXVtbl is and what an 
}IY;                          //IYVtbl is.  They are arrays of function pointers 'templated', so to speak, within a containing

struct IXVtbl                                                   //struct.  No such struct has as yet been instantiated, although
{                                                               //we have instantiated two other structures named IX and IY that 
 HRESULT (__stdcall* QueryInterface) (IX*, const IID*, void**); //contain each a pointer member which will point to one of these 
 ULONG   (__stdcall* AddRef)         (IX*                    ); //structures once we actually instantiate one.  And the confusion 
 ULONG   (__stdcall* Release)        (IX*                    ); //continues below left where we amass into yet another containing 
 HRESULT (__stdcall* Fx1)            (IX*, int               ); //structure named CB pointers to these IXVtbl and IYVtbl 
 HRESULT (__stdcall* Fx2)            (IX*, int               ); //arrays/structures.  Additionally, an integer counter variable
};                                                              

struct IYVtbl                                                   //m_cRef has been added to what at this point now constitutes a 
{                                                               //class.  It should be clear though that at some point we're 
 HRESULT (__stdcall* QueryInterface) (IY*, const IID*, void**); //going to have to cease defining pointers to 'things' and start 
 ULONG   (__stdcall* AddRef)         (IY*                    ); //defining the 'things' themselves.  That is exactly what is 
 ULONG   (__stdcall* Release)        (IY*                    ); //going on just below left with IX_QueryInterface where we are 
 HRESULT (__stdcall* Fy1)            (IY*, int               ); //defining a function that actually takes parameters, does 
 HRESULT (__stdcall* Fy2)            (IY*, int               ); //something meaningful, and returns something.  It is a 
};                                                              //fundamental rule of COM that every interface, i.e., vtable, 


typedef struct     //must contain pointers to QueryInterface(), AddRef(), and Release() as its 1st three member functions.  In 
{                  //this sence, every COM object is a polymorph of IUnknown, or polymorphic in IUnknown.  The IUnknown functions
 IXVtbl* lpIXVtbl; //of every interface represent system services in that they provide to an object the capability of being  
 IYVtbl* lpIYVtbl; //queried for additional interfaces, of being reference counted, and being released from memory when no longer
 int     m_cRef;   //needed.  Since this particular object, i.e., CB, alias Class B, contains two interfaces, we're going to have
}CB;               //to implement ten functions, pointers to which will be stored in their respective VTable. You will note the...   


HRESULT __stdcall IX_QueryInterface(IX* this, const IID* iid, void** ppv)
{                                     
 *ppv=0;                               //1st function pointer in IXVtbl is QueryInterface, and that function pointer will be 
 if(!memcmp(iid,&IID_IUnknown,16))     //initialized on the Dll's load with the address of IX_QueryInterface().  If you want to   
 {                                     //jump ahead and see where this is done look below about 160 lines just before the 
    printf("IX_QueryInterface() ");    //IClassFactory code and you'll see where all the IXVtbl and IYVtbl function pointers are
    printf("For IUnknown!\t\t");       //initialized to the addresses of the actual IX and IY interface functions just left and 
    printf("this=%u\n",this);          //down.  Note that that particular code looks something like array initialization code in
    *ppv=this;                         //C, but structs can be initialized the same way, and that is what you are seeing there.
 }                                     //Note that the syntax - const IXVtbl IX_Vtbl and const IYVtbl IY_Vtbl makes IX_Vtbl and
 if(!memcmp(iid, &IID_IX, 16))         //IY_Vtbl global to this app/dll and constant.  And of course, once the compiler assigns
 {                                     //the addresses of these ten interface functions to their respective VTable, these addresses
    printf("IX_QueryInterface() ");    //will not change after Dll load.  One thing you may note that this particular app written
    printf("For IX!\t\t");             //in C makes clear as opposed to my C++ derivation in Tutorial #1, is that each interface
    printf("this=%u\n",this);          //has a seperate implementation of the IUnknown functions.  Examine for a second just above 
    *ppv=this;                         //the definition of the IXVtbl and the IYVtbl. They each contain function pointers named
 }                                     //QueryInterface, AddRef, and Release, in addition to their respective Fx1/Fy1 and fx2/Fy2
 if(!memcmp(iid, &IID_IY, 16))         //function pointers.  However, the QueryInterface pointer in the IXVtbl won't point to the
 {                                     //same function as the QueryInterface pointer in the IYVtbl.  The QueryInterface pointer
    IY* pIY=(IY*)this;                 //member of the IXVtbl will point to the address of the function just left of these words
    pIY++;                             //I'm now writting, which is of course IX_QueryInterface.  
    *ppv=pIY;
    printf("IX_QueryInterface() ");    //Further note that the first interface in a class is somewhat special in that, if a client 
    printf("For IY!\t\t");             //creates the COM object and asks for the IUnknown interface, the pointer returned will
    printf("this=%u\n",this);          //simultaneously be a pointer to the first VTable defined in the class.  That is why when
    pIY->lpVtbl->AddRef(pIY);          //you examine the if statements just left you'll see that the IX interface pointer input
    return S_OK;                       //into IX_QueryInterface as the first parameter is simply output through the last output
 }                                     //parameter (a void**) unchanged.  Note what happens though when the iid passed in is for 
 if(*ppv)                              //the IY interface.  In that case an increment operation ( pIY++ ) is performed on the 
 {                                     //address passed in through the this pointer so as to return the pointer in the next Vtbl
    this->lpVtbl->AddRef(this);        //Ptr slot in CB's memory.
    return S_OK;  
 }                                     //Another interesting point you may note if you compare these IX_QueryInterface() or
 printf("Called IX_QueryInterface()"); //IY_QueryInterface() functions with what we had in tutorial #1's class CA written in C++
 printf("\tthis=%u\n",this);           //is that we are using the memcmp() C Runtime function to determine what IID was passed
                                       //in whereas in C++ we simply tested for equality like so - if(riid==IID_IUnknown).  Well,
 return(E_NOINTERFACE);                //that niceity was a gift of C++ with its operator overloading functionality which auto- 
}                                      //matically causes a function to be called in objbase.h that does the actual comparisons...
                                  

ULONG __stdcall IX_AddRef(IX* this)    //using memcmp.  Another issue one must face in writting this code in C as opposed to C++
{                                      //is that the AddRef() and Release() operations sometimes have a nasty twist to them, not
 printf("Called IX_AddRef()!\t\t");    //so much with the IX functions (or the first interface in the class) but with those that
 printf("this=%u\n",this);             //come after (here IY).  Actually, the IX situation is even a bit nasty.  Just look at   
 return(++((CB*)this)->m_cRef);        //that ugly 'return' just left for IX_AddRef()!  Here though its just a casting issue as
}                                      //the IX* ( IX pointer ) passed in through the parameter list, while holding the correct... 


ULONG __stdcall IX_Release(IX* this)   //address of class CB, is of the incorrect type and so has to be cast to a CB* before the
{                                      //m_cRef reference counting variable can be accessed.  Even in the PowerBASIC app i takes
 if(--((CB*)this)->m_cRef == 0)        //some strange machinations to accomplish it as one must first declare a CB Ptr variable
 {                                      
     CoTaskMemFree(this);              //Local pCB As CB Ptr'
     InterlockedDecrement(&g_lObjs);   //pCB=this
     printf("Called IX_Release() ");   //Incr @pCB.m_cRef
     printf("And CB Was Deleted!\n");  
     return(0);                        //and access the reference counter through it.  Jump below for discussion of the IY
 }                                     //situation.
 printf("Called IX_Release()!\t\t");
 printf("this=%u\n",this);

 return(((CB*)this)->m_cRef);
}


HRESULT __stdcall Fx1(IX* this, int iNum)
{
 printf("Called Fx1()  :  iNum = %u\n",
 iNum);                            
 return S_OK;
}


HRESULT __stdcall Fx2(IX* this, int iNum)
{
 printf("Called Fx2()  :  iNum = %u\n",
 iNum);
 return S_OK;
}


HRESULT __stdcall IY_QueryInterface(IY* this, const IID* iid, void** ppv)
{
 *ppv=0;
 if(!memcmp(iid,&IID_IUnknown,16))     //When any of the IY interface functions are called a pointer to the IY interface will
 {                                     //be coming in through the first parameter.  From my discussion in the tutorial associated
    IX* pIX=(IX*)this;                 //with this code you saw that the IY VTable pointer will be occupying the 4th through 
    pIX--;                             //7th bytes of the 12 byte memory block that constitutes class CB.  The significance of
    printf("IY_QueryInterface() ");    //this of course is that to either return an IX or IUnknown pointer in IY_QueryInterface
    printf("For IUnknown!\t\t");       //or access the m_cRef parameter back in class CB, a decrement operation is going to have 
    printf("this=%u\n",this);          //to be performed on the incoming this pointer, as well as the necessary casting.  That is
    *ppv=pIX;                          //what you see going on just left.  In the case of the IY interface, of course, all that
    pIX->lpVtbl->AddRef(pIX);          //needs done is to output the same number that came in, i.e., *ppv=this.
    return S_OK;    
 }
 if(!memcmp(iid,&IID_IX,16))
 {
    IX* pIX=(IX*)this;

    pIX--;
    printf("IY_QueryInterface() For IX!\tthis=%u\n",this);
    *ppv=pIX;
    pIX->lpVtbl->AddRef(pIX);
    return S_OK;
 }
 if(!memcmp(iid,&IID_IY,16))
 {
    this->lpVtbl->AddRef(this);
    printf("IY_QueryInterface() For IY!\tthis=%u\n",this);
    *ppv=this;
    return S_OK;
 }
 printf("Called IY_QueryInterface()\tthis=%u\n",this);  
 
 return(E_NOINTERFACE);
}


ULONG __stdcall IY_AddRef(IY* this)
{
 printf("Called IY_AddRef()!\t\tthis=%u\n",this);
 this--;
 return(++((CB*)this)->m_cRef);
}


ULONG __stdcall IY_Release(IY* this)
{
 CB* pCB;

 printf("Called IY_Release()\t\tthis=%u\n",this); 
 this--;
 pCB=(CB*)this;
 pCB->m_cRef--;
 if(!pCB->m_cRef)
 {
    CoTaskMemFree(this);
    InterlockedDecrement(&g_lObjs);
    printf("Called IY_Release() And CB Was Deleted!\n");
    return(0);
 }
 
 return pCB->m_cRef;
}


HRESULT __stdcall Fy1(IY* this, int iNum)
{
 printf("Called Fy1()  :  iNum = %u\n",iNum);
 return S_OK;
}





HRESULT __stdcall Fy2(IY* this, int iNum)
{
 printf("Called Fy2()  :  iNum = %u\n",iNum);
 return S_OK;
}


const IXVtbl IX_Vtbl=  //This here is of critical importance to understand.  Previously near the top of 
{                      //this code the IXVtbl and IYVtbl structures were defined.  They were defined as
 IX_QueryInterface,    //containing nothing but function pointers.  Here variables of these structure
 IX_AddRef,            //types are being created (which will cause a memory allocation) and the pointer
 IX_Release,           //members are being initialized to the addresses of the above IX and IY interface
 Fx1,                  //member functions.  This looks like an array allocation in C but the C language
 Fx2                   //allows structures to be initialized like this too.  In PowerBASIC we used the
};                     //CodePtr function to set the addresses in the IX/IYVtbl type variables, but in...


const IYVtbl IY_Vtbl=  //C the unadorned function name is all that is needed for the compiler to return
{                      //the address represented by such a symbol or program 'token'.  What you see right 
 IY_QueryInterface,    //below here too is also important as this is the end of the necessary setup for 
 IY_AddRef,            //the IX and IY interfaces, and the code related to CB's Class Factory implemen-
 IY_Release,           //tation is just starting with the declaration of the CBClassFactory variable
 Fy1,                  //of type IClassFactory.  This type is defined in Unknwn.h and that is included
 Fy2                   //by default.  Following the declaration of CBClassFactory are implementations of
};                     //the five IClassFactory functions, and the addresses of these functions are set


IClassFactory  CBClassFactory;  //in an analogous manner to the members within the IClassFactory_Vtbl...


HRESULT __stdcall CBClassFactory_QueryInterface(IClassFactory* this, REFIID riid, void** ppv)
{
 if(IsEqualIID(riid,&IID_IUnknown) || IsEqualIID(riid,&IID_IClassFactory))
 {
    this->lpVtbl->AddRef(this); //structure just below CBClassFactory_LockServer().  Note that here in
    *ppv=this;                  //CBClassFactory_QueryInterface(), if IID_IUnknown or IID_IClassFactory
    return(NOERROR);            //is passed in through the riid parameter, the function simply returns
 }                              //the IClassFactory* passed in through the first parameter through the 
 *ppv=0;                        //last parameter, i.e., void** ppv.  The way I say or think of these
                                //sorts of operations mentally, that is, *ppv=this, is to say, 'what's
 return(E_NOINTERFACE);         //stored at ppv is "this".
}


ULONG __stdcall CBClassFactory_AddRef(IClassFactory* this)  //InterlockedIncrement() and InterlockedDecrement()
{                                                           //are typically used to increment/decrement reference
 InterlockedIncrement(&g_lObjs);                            //counting variables because they are thread safe.
 return(1);
}


ULONG __stdcall CBClassFactory_Release(IClassFactory* this)
{
 return(InterlockedDecrement(&g_lObjs));
}


HRESULT __stdcall CBClassFactory_CreateInstance(IClassFactory* this, IUnknown* pUnkOuter, REFIID iid, void** ppv)
{
 CB* pCB=NULL;                                             //This is a critically important function in COM that actually 
 HRESULT  hr;                                              //creates the class and causes a VTable pointer to some specific
	                                                       //class interface (identified by the REFIID parameter) to be   
 *ppv=0;                                                   //returned to the client.  This particular implementation first
 if(pUnkOuter)                                             //reassures itself that pUnkOuter parameter is NULL, because 
    hr=CLASS_E_NOAGGREGATION;                              //class CB doesn't support aggregation.  If the 2nd parameter
 else                                                      //is NULL CoTaskMemAlloc() is used to request 12 bytes of memory
 {                                                         //to store two VTable Pointers and a reference counting variable
    pCB=(CB*)CoTaskMemAlloc(sizeof(CB));                   //named m_cRef.  Then constant structures of type IXVtbl and 
    if(pCB)                                                //IYVtbl (IX_Vtbl and IY_Vtbl), whose members just above we filled 
    {                                                      //in with the addresses of the various IX and IY interface 
       pCB->lpIXVtbl=(IXVtbl*)&IX_Vtbl;                    //member functions, have their respective addresses set or plugged
       pCB->lpIYVtbl=(IYVtbl*)&IY_Vtbl;                    //into the VTable Pointer members of the containing CB class.  So
       pCB->m_cRef=0;                                      //what's stored in the first four of CB's twelve bytes is the
       hr=pCB->lpIXVtbl->QueryInterface((IX*)pCB,iid,ppv); //address of the IX VTable, and in the second four bytes the
       if(SUCCEEDED(hr))                                   //address of the IY VTable (on 32 bit systems).  Next the REFIID 
          InterlockedIncrement(&g_lObjs);                  //passed in through the 3rd parameter is passed to IX_QueryInterface
       else                                                //and if that REFIID is IID_IUnknown, IID_IX, or IID_IY, then
          CoTaskMemFree(pCB);                              //the function will succeed and the address passed into
    }                                                      //CBClassFactory_CreateInstance() in the last parameter will have
    else                                                   //an interface pointer address stuffed into it as the final result
       hr=E_OUTOFMEMORY;                                   //of all these various machinations.  And it is important to
 }                                                         //realize the address passed into this function in that last
                                                           //parameter is back in the client app where, in PowerBASIC code
 return(hr);                                               //there may be a variable declaration such as Local pIX As IX or
}                                                          //in C or C++ something like IX* pIX=NULL; 


HRESULT __stdcall CBClassFactory_LockServer(IClassFactory* this, BOOL flock)
{
 if(flock)
    InterlockedIncrement(&g_lLocks);                       //LockServer is used to prevent the Dll from being unloaded by 
 else                                                      //the operating system if there are no class objects presently  
    InterlockedDecrement(&g_lLocks);                       //existing, i.e., the m_cRef member became 0 for all objects,
                                                           //and g_lObjs is also 0.  If g_lLocks is positive, DllCanUnloadNow(),
 return(NOERROR);                                          //which is called periodically by the operating system, will
}                                                          //always return FALSE.


const IClassFactoryVtbl IClassFactory_Vtbl=                //The addresses of the five functions of the IClassFactory interface 
{                                                          //are plugged into its VTable here.  If you look below about 45 lines 
 CBClassFactory_QueryInterface,                            //you'll see in DllMain() where the CBClassFactory.lpVtbl member is
 CBClassFactory_AddRef,                                    //initialized with the runtime address of this structure defined in
 CBClassFactory_Release,                                   //Unknwn.h.  In the PowerBASIC Dll's code we defined the structure
 CBClassFactory_CreateInstance,                            //ourselves.
 CBClassFactory_LockServer
};


HRESULT __stdcall DllGetClassObject(REFCLSID objGuid, REFIID facGuid, void** ppv)
{
 HRESULT hr;

 if(!memcmp(objGuid,&CLSID_CB,16))
    CBClassFactory_QueryInterface(&CBClassFactory, facGuid, ppv);
 else
 {
    *ppv=0;
    hr=CLASS_E_CLASSNOTAVAILABLE;
 }
 
 return(hr);
}


HRESULT __stdcall DllCanUnloadNow()
{
 if(g_lObjs||g_lLocks)
    return S_FALSE;
 else
    return S_OK;
}


STDAPI DllRegisterServer()
{                         
 return RegisterServer(g_hModule,&CLSID_CB,g_szFriendlyName,g_szVerIndProgID,g_szProgID);                       
}                         


STDAPI DllUnregisterServer()
{                           
 return UnregisterServer(&CLSID_CB,g_szVerIndProgID,g_szProgID);
}  


BOOL __stdcall DllMain(HINSTANCE hInstance, DWORD fdwReason, LPVOID lpvReserved)
{
 if(fdwReason==DLL_PROCESS_ATTACH)
 {
    g_hModule=hInstance;
    CBClassFactory.lpVtbl = (IClassFactoryVtbl*)&IClassFactory_Vtbl;
    DisableThreadLibraryCalls(hInstance);
 }

 return(TRUE);
}


/*
//debug Version
#include     <objbase.h>       //This program is exactly similiar to the CA COM object I developed in C++ in my 1st COM
#include     <stdio.h>         //Object Memory Tutorial.  But in this tutorial we are going a bit lower level and will
#include     "Registry.h"      //use plain C and plain PowerBASIC, i.e., PowerBASIC without Version 9's functionality.

const      CLSID CLSID_CB      ={0x20000000,0x0000,0x0000,{0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x10}}; 
const      IID   IID_IX        ={0x20000000,0x0000,0x0000,{0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x11}};
const      IID   IID_IY        ={0x20000000,0x0000,0x0000,{0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x12}};
HINSTANCE  g_hModule           = NULL;              //Store dll instance handle
const char g_szFriendlyName[]  = "Com Object CB";   //Store friendly name of component
const char g_szVerIndProgID[]  = "ComObject.CB";    //Store Version Independent ProgID
const char g_szProgID[]        = "ComObject.CB.1";  //Store Versioned Program ID.
long       g_lObjs=0;
long       g_lLocks=0;

typedef struct IXVtbl IXVtbl; //All these two lines do is specify the existance of a struct variable named IXVtbl or IYVtbl 
typedef struct IYVtbl IYVtbl; //respectively without actually defining what these structures are or allocating any storage for  

typedef interface IX          //same.  The typedef just left though does both define and allocate storage for another entity         
{                             //that contains a pointer to one of these as yet undefined entities, i.e., an IXVtbl or an IYVtbl.
 const IXVtbl* lpVtbl;        //C allows this because while it may not as yet know what an IXVtbl or IYVtbl is, it does know how 
}IX;                          //big a pointer to one of them would be.  The critical lines immediately below though finally 

typedef interface IY          //reveal just what an IXVtbl and an IYVtbl are.  We see that an IXVtbl and an IYVtbl contain 
{                             //function pointers.  So lets take stock at this point of just what has been defined only, and    
 const IYVtbl* lpVtbl;        //what has actualy been allocated.  Its not hard to do. We do know what an IXVtbl is and what an 
}IY;                          //IYVtbl is.  They are arrays of function pointers 'templated', so to speak, within a containing

struct IXVtbl                                                   //struct.  No such struct has as yet been instantiated, although
{                                                               //we have instantiated two other structures named IX and IY that 
 HRESULT (__stdcall* QueryInterface) (IX*, const IID*, void**); //contain each a pointer member which will point to one of these 
 ULONG   (__stdcall* AddRef)         (IX*                    ); //structures once we actually instantiate one.  And the confusion 
 ULONG   (__stdcall* Release)        (IX*                    ); //continues below left where we amass into yet another containing 
 HRESULT (__stdcall* Fx1)            (IX*, int               ); //structure named CB pointers to these IXVtbl and IYVtbl 
 HRESULT (__stdcall* Fx2)            (IX*, int               ); //arrays/structures.  Additionally, an integer counter variable
};                                                              

struct IYVtbl                                                   //m_cRef has been added to what at this point now constitutes a 
{                                                               //class.  It should be clear though that at some point we're 
 HRESULT (__stdcall* QueryInterface) (IY*, const IID*, void**); //going to have to cease defining pointers to 'things' and start 
 ULONG   (__stdcall* AddRef)         (IY*                    ); //defining the 'things' themselves.  That is exactly what is 
 ULONG   (__stdcall* Release)        (IY*                    ); //going on just below left with IX_QueryInterface where we are 
 HRESULT (__stdcall* Fy1)            (IY*, int               ); //defining a function that actually takes parameters, does 
 HRESULT (__stdcall* Fy2)            (IY*, int               ); //something meaningful, and returns something.  It is a 
};                                                              //fundamental rule of COM that every interface, i.e., vtable, 


typedef struct     //must contain pointers to QueryInterface(), AddRef(), and Release() as its 1st three member functions.  In 
{                  //this sence, every COM object is a polymorph of IUnknown, or polymorphic in IUnknown.  The IUnknown functions
 IXVtbl* lpIXVtbl; //of every interface represent system services in that they provide to an object the capability of being  
 IYVtbl* lpIYVtbl; //queried for additional interfaces, of being reference counted, and being released from memory when no longer
 int     m_cRef;   //needed.  Since this particular object, i.e., CB, alias Class B, contains two interfaces, we're going to have
}CB;               //to implement ten functions, pointers to which will be stored in their respective VTable. You will note the...   


HRESULT __stdcall IX_QueryInterface(IX* this, const IID* iid, void** ppv)
{                                     
 *ppv=0;                               //1st function pointer in IXVtbl is QueryInterface, and that function pointer will be 
 if(!memcmp(iid,&IID_IUnknown,16))     //initialized on the Dll's load with the address of IX_QueryInterface().  If you want to   
 {                                     //jump ahead and see where this is done look below about 160 lines just before the 
    printf("IX_QueryInterface() ");    //IClassFactory code and you'll see where all the IXVtbl and IYVtbl function pointers are
    printf("For IUnknown!\t\t");       //initialized to the addresses of the actual IX and IY interface functions just left and 
    printf("this=%u\n",this);          //down.  Note that that particular code looks something like array initialization code in
    *ppv=this;                         //C, but structs can be initialized the same way, and that is what you are seeing there.
 }                                     //Note that the syntax - const IXVtbl IX_Vtbl and const IYVtbl IY_Vtbl makes IX_Vtbl and
 if(!memcmp(iid, &IID_IX, 16))         //IY_Vtbl global to this app/dll and constant.  And of course, once the compiler assigns
 {                                     //the addresses of these ten interface functions to their respective VTable, these addresses
    printf("IX_QueryInterface() ");    //will not change after Dll load.  One thing you may note that this particular app written
    printf("For IX!\t\t");             //in C makes clear as opposed to my C++ derivation in Tutorial #1, is that each interface
    printf("this=%u\n",this);          //has a seperate implementation of the IUnknown functions.  Examine for a second just above 
    *ppv=this;                         //the definition of the IXVtbl and the IYVtbl. They each contain function pointers named
 }                                     //QueryInterface, AddRef, and Release, in addition to their respective Fx1/Fy1 and fx2/Fy2
 if(!memcmp(iid, &IID_IY, 16))         //function pointers.  However, the QueryInterface pointer in the IXVtbl won't point to the
 {                                     //same function as the QueryInterface pointer in the IYVtbl.  The QueryInterface pointer
    IY* pIY=(IY*)this;                 //member of the IXVtbl will point to the address of the function just left of these words
    pIY++;                             //I'm now writting, which is of course IX_QueryInterface.  
    *ppv=pIY;
    printf("IX_QueryInterface() ");    //Further note that the first interface in a class is somewhat special in that, if a client 
    printf("For IY!\t\t");             //creates the COM object and asks for the IUnknown interface, the pointer returned will
    printf("this=%u\n",this);          //simultaneously be a pointer to the first VTable defined in the class.  That is why when
    pIY->lpVtbl->AddRef(pIY);          //you examine the if statements just left you'll see that the IX interface pointer input
    return S_OK;                       //into IX_QueryInterface as the first parameter is simply output through the last output
 }                                     //parameter (a void**) unchanged.  Note what happens though when the iid passed in is for 
 if(*ppv)                              //the IY interface.  In that case an increment operation ( pIY++ ) is performed on the 
 {                                     //address passed in through the this pointer so as to return the pointer in the next Vtbl
    this->lpVtbl->AddRef(this);        //Ptr slot in CB's memory.
    return S_OK;  
 }                                     //Another interesting point you may note if you compare these IX_QueryInterface() or
 printf("Called IX_QueryInterface()"); //IY_QueryInterface() functions with what we had in tutorial #1's class CA written in C++
 printf("\tthis=%u\n",this);           //is that we are using the memcmp() C Runtime function to determine what IID was passed
                                       //in whereas in C++ we simply tested for equality like so - if(riid==IID_IUnknown).  Well,
 return(E_NOINTERFACE);                //that niceity was a gift of C++ with its operator overloading functionality which auto- 
}                                      //matically causes a function to be called in objbase.h that does the actual comparisons...
                                  

ULONG __stdcall IX_AddRef(IX* this)    //using memcmp.  Another issue one must face in writting this code in C as opposed to C++
{                                      //is that the AddRef() and Release() operations sometimes have a nasty twist to them, not
 printf("Called IX_AddRef()!\t\t");    //so much with the IX functions (or the first interface in the class) but with those that
 printf("this=%u\n",this);             //come after (here IY).  Actually, the IX situation is even a bit nasty.  Just look at   
 return(++((CB*)this)->m_cRef);        //that ugly 'return' just left for IX_AddRef()!  Here though its just a casting issue as
}                                      //the IX* ( IX pointer ) passed in through the parameter list, while holding the correct... 


ULONG __stdcall IX_Release(IX* this)   //address of class CB, is of the incorrect type and so has to be cast to a CB* before the
{                                      //m_cRef reference counting variable can be accessed.  Even in the PowerBASIC app i takes
 if(--((CB*)this)->m_cRef == 0)        //some strange machinations to accomplish it as one must first declare a CB Ptr variable
 {                                      
     CoTaskMemFree(this);              //Local pCB As CB Ptr'
     InterlockedDecrement(&g_lObjs);   //pCB=this
     printf("Called IX_Release() ");   //Incr @pCB.m_cRef
     printf("And CB Was Deleted!\n");  
     return(0);                        //and access the reference counter through it.  Jump below for discussion of the IY
 }                                     //situation.
 printf("Called IX_Release()!\t\t");
 printf("this=%u\n",this);

 return(((CB*)this)->m_cRef);
}


HRESULT __stdcall Fx1(IX* this, int iNum)
{
 printf("Called Fx1()  :  iNum = %u\n",
 iNum);                            
 return S_OK;
}


HRESULT __stdcall Fx2(IX* this, int iNum)
{
 printf("Called Fx2()  :  iNum = %u\n",
 iNum);
 return S_OK;
}


HRESULT __stdcall IY_QueryInterface(IY* this, const IID* iid, void** ppv)
{
 *ppv=0;
 if(!memcmp(iid,&IID_IUnknown,16))     //When any of the IY interface functions are called a pointer to the IY interface will
 {                                     //be coming in through the first parameter.  From my discussion in the tutorial associated
    IX* pIX=(IX*)this;                 //with this code you saw that the IY VTable pointer will be occupying the 4th through 
    pIX--;                             //7th bytes of the 12 byte memory block that constitutes class CB.  The significance of
    printf("IY_QueryInterface() ");    //this of course is that to either return an IX or IUnknown pointer in IY_QueryInterface
    printf("For IUnknown!\t\t");       //or access the m_cRef parameter back in class CB, a decrement operation is going to have 
    printf("this=%u\n",this);          //to be performed on the incoming this pointer, as well as the necessary casting.  That is
    *ppv=pIX;                          //what you see going on just left.  In the case of the IY interface, of course, all that
    pIX->lpVtbl->AddRef(pIX);          //needs done is to output the same number that came in, i.e., *ppv=this.
    return S_OK;    
 }
 if(!memcmp(iid,&IID_IX,16))
 {
    IX* pIX=(IX*)this;

    pIX--;
    printf("IY_QueryInterface() For IX!\tthis=%u\n",this);
    *ppv=pIX;
    pIX->lpVtbl->AddRef(pIX);
    return S_OK;
 }
 if(!memcmp(iid,&IID_IY,16))
 {
    this->lpVtbl->AddRef(this);
    printf("IY_QueryInterface() For IY!\tthis=%u\n",this);
    *ppv=this;
    return S_OK;
 }
 printf("Called IY_QueryInterface()\tthis=%u\n",this);  
 
 return(E_NOINTERFACE);
}


ULONG __stdcall IY_AddRef(IY* this)
{
 printf("Called IY_AddRef()!\t\tthis=%u\n",this);
 this--;
 return(++((CB*)this)->m_cRef);
}


ULONG __stdcall IY_Release(IY* this)
{
 CB* pCB;

 printf("Called IY_Release()\t\tthis=%u\n",this); 
 this--;
 pCB=(CB*)this;
 pCB->m_cRef--;
 if(!pCB->m_cRef)
 {
    CoTaskMemFree(this);
    InterlockedDecrement(&g_lObjs);
    printf("Called IY_Release() And CB Was Deleted!\n");
    return(0);
 }
 
 return pCB->m_cRef;
}


HRESULT __stdcall Fy1(IY* this, int iNum)
{
 printf("Called Fy1()  :  iNum = %u\n",iNum);
 return S_OK;
}


HRESULT __stdcall Fy2(IY* this, int iNum)
{
 printf("Called Fy2()  :  iNum = %u\n",iNum);
 return S_OK;
}


const IXVtbl IX_Vtbl=  //This here is of critical importance to understand.  Previously near the top of 
{                      //this code the IXVtbl and IYVtbl structures were defined.  They were defined as
 IX_QueryInterface,    //containing nothing but function pointers.  Here variables of these structure
 IX_AddRef,            //types are being created (which will cause a memory allocation) and the pointer
 IX_Release,           //members are being initialized to the addresses of the above IX and IY interface
 Fx1,                  //member functions.  This looks like an array allocation in C but the C language
 Fx2                   //allows structures to be initialized like this too.  In PowerBASIC we used the
};                     //CodePtr function to set the addresses in the IX/IYVtbl type variables, but in...


const IYVtbl IY_Vtbl=  //C the unadorned function name is all that is needed for the compiler to return
{                      //the address represented by such a symbol or program 'token'.  What you see right 
 IY_QueryInterface,    //below here too is also important as this is the end of the necessary setup for 
 IY_AddRef,            //the IX and IY interfaces, and the code related to CB's Class Factory implemen-
 IY_Release,           //tation is just starting with the declaration of the CBClassFactory variable
 Fy1,                  //of type IClassFactory.  This type is defined in Unknwn.h and that is included
 Fy2                   //by default.  Following the declaration of CBClassFactory are implementations of
};                     //the five IClassFactory functions, and the addresses of these functions are set


IClassFactory  CBClassFactory;


HRESULT __stdcall CBClassFactory_QueryInterface(IClassFactory* this, REFIID factoryGuid, void** ppv)
{
 printf("\nEntering CBClassFactory_QueryInterface()\n");
 printf("    this = %u\n",this);
 if(IsEqualIID(factoryGuid,&IID_IUnknown) || IsEqualIID(factoryGuid,&IID_IClassFactory))
 {
    this->lpVtbl->AddRef(this);
    *ppv=this;
    printf("  this=%u\n",this);
    printf("  *ppv=%u\n",*ppv);
    printf("\nLeaving CBClassFactory_QueryInterface()\n");
    return(NOERROR);
 }
 *ppv=0;
 printf("Leaving CBClassFactory_QueryInterface() And ppv=0\n\n");

 return(E_NOINTERFACE);
}


ULONG __stdcall CBClassFactory_AddRef(IClassFactory* this)
{
 printf("Entering CBClassFactory_AddRef()\n");
 InterlockedIncrement(&g_lObjs);
 printf("  g_lObjs=%u\n",g_lObjs);
 printf("Leaving CBClassFactory_AddRef()\n\n");

 return(g_lObjs);
}


ULONG __stdcall CBClassFactory_Release(IClassFactory* this)
{
 printf("Entering CBClassFactory_Release()\n");
 InterlockedDecrement(&g_lObjs);
 printf("  g_lObjs=%u\n",g_lObjs);
 printf("Leaving CBClassFactory_Release()\n\n");

 return(g_lObjs);
}


HRESULT __stdcall CBClassFactory_CreateInstance(IClassFactory* this, IUnknown* pUnkOuter, REFIID iid, void** ppv)
{
 CB* pCB=NULL;
 HRESULT  hr;
	
 printf("\nEntering CBClassFactory_CreateInstance()\n");
 *ppv=0;
 if(pUnkOuter)
    hr=CLASS_E_NOAGGREGATION;
 else
 {
    pCB=(CB*)CoTaskMemAlloc(sizeof(CB));
    if(pCB)
    {
       printf("pCB=%u\n",pCB);
       pCB->lpIXVtbl=(IXVtbl*)&IX_Vtbl,  pCB->lpIYVtbl=(IYVtbl*)&IY_Vtbl;
       printf("pCB->lpIXVtbl=%u\n",pCB->lpIXVtbl);
       printf("pCB->lpIYVtbl=%u\n\n",pCB->lpIYVtbl);

       pCB->m_cRef=0; 
       hr=pCB->lpIXVtbl->QueryInterface((IX*)pCB,iid,ppv);
       if(SUCCEEDED(hr))
       {
          printf("QueryInterface(IX) Succeeded And *ppv=%u\n",*ppv);
          InterlockedIncrement(&g_lObjs);
          printf("g_lObjs=%u\n\n",g_lObjs);
          printf("%u\t%u\n",&pCB->lpIXVtbl,pCB->lpIXVtbl);
          printf("%u\t%u\n",&pCB->lpIYVtbl,pCB->lpIYVtbl);
       }
       else
       {
          printf("QueryInterface(IX) Failed And ppv=%u.  pCB Will Be Freed!\n",ppv);
          CoTaskMemFree(pCB);
       }
    }
    else
       hr=E_OUTOFMEMORY;
 }
 printf("Leaving CBClassFactory_CreateInstance()\n");
 
 return(hr);
}


HRESULT __stdcall CBClassFactory_LockServer(IClassFactory* this, BOOL flock)
{
 if(flock)
    InterlockedIncrement(&g_lLocks);
 else 
    InterlockedDecrement(&g_lLocks);

 return(NOERROR);
}


const IClassFactoryVtbl IClassFactory_Vtbl=
{
 CBClassFactory_QueryInterface,
 CBClassFactory_AddRef,
 CBClassFactory_Release,
 CBClassFactory_CreateInstance,
 CBClassFactory_LockServer
};


HRESULT __stdcall DllGetClassObject(REFCLSID objGuid, REFIID facGuid, void** ppv)
{
 HRESULT hr;

 printf("\nEntering DLLGetClassObject()\n");
 printf("  Got Inside If So We're Requesting Either CLSID_CB Or IID_IClassFactory!\n");
 printf("  &CBClassFactory = %u\n",&CBClassFactory);
 if(!memcmp(objGuid,&CLSID_CB,16))
 {
    hr=CBClassFactory_QueryInterface(&CBClassFactory, facGuid, ppv);
    printf("*ppv=%u\n",*ppv);
 }
 else
 {
    *ppv=0;
    hr=CLASS_E_CLASSNOTAVAILABLE;
    printf("CLSID_CB Wasn't Passed In!  Not Good!\n");
 }
 printf("Leaving DllGetClassObject()\n\n");
 
 return(hr);
}


HRESULT __stdcall DllCanUnloadNow()
{
 printf("Entering DllCanUnloadNow()\n");
 printf("  g_lObjs  = %u\n",g_lObjs);
 printf("  g_lLocks = %u\n",g_lLocks);
 printf("Leaving DllCanUnloadNow()\n");
 if(g_lObjs||g_lLocks)
    return S_FALSE;
 else
    return S_OK;
}


STDAPI DllRegisterServer()
{                         
 return RegisterServer(g_hModule,&CLSID_CB,g_szFriendlyName,g_szVerIndProgID,g_szProgID);                       
}                         


STDAPI DllUnregisterServer()
{                           
 return UnregisterServer(&CLSID_CB,g_szVerIndProgID,g_szProgID);
}  


BOOL __stdcall DllMain(HINSTANCE hInstance, DWORD fdwReason, LPVOID lpvReserved)
{
 if(fdwReason==DLL_PROCESS_ATTACH)
 {
    printf("Received DLL_PROCESS_ATTACH\n");
    g_hModule=hInstance;
    CBClassFactory.lpVtbl = (IClassFactoryVtbl*)&IClassFactory_Vtbl;
    DisableThreadLibraryCalls(hInstance);
 }
 if(fdwReason==DLL_PROCESS_DETACH)
    printf("\nReceived DLL_PROCESS_DETACH\n");

 return(TRUE);
}
*/
